/*
**************************************************************************************************
*                               Embedded Systems Building Blocks
*                             PS/2 Keyboard 104 keys, NumPad driver
*
*                    Original obtained from www.picant.com, C2C complier web site
*                        by Julio Cesar Silva Briano, modified by John Leung
*                                  extended by Klemens Schatka
*
*                                     original text follows
*
* Filename 		: ps2key.c
* Programmer 	: John Leung, www.TechToys.com.hk under Downloads Section.
* Version		: Version 1.1 - 17th June 2004
* Remarks		: Consult www.beyong.org for PS/2 keyboard protocol
**************************************************************************************************
*											DESCRIPTION
*		This is a PS/2 keyboard driver for standard 101/104 keyboard or Numerical Keypad
*	 			Application software must include PS2Key.h and PS2ScanCode.h
*							PS2Key.h contains hardware port configuration
*			PS2ScanCode.h configures if it is a complete 101 keyboard, or just the Numpad
*	Two lookup tables PS2KeyUnshifted[] and PS2KeyShifted[] map the PS2Key_Code to ASCII characters
* If it is going to be the Numpad ONLY, PS2KeyUnshifted[] and PS2KeyShifted[] can be undefined to save ROM
*					Application must declare the followings (see ps2key.h)
*	PS2key.c is PS/2 keyboard driver to handle ordinary PS/2 or AT-style keyboard with mcu
*	Global variable PS2Key_Code returns the raw key code of a key
*	Key Buffer is defined by constant PS2KEY_BUF_SIZE (in PS2Key.h)
*
*	PS2Key_Clk			The port address of the PS/2 CLK signal
*	PS2Key_Data			The port address of the PS/2 Data signal
*	PS2Key_BUF_SIZE 	Size of the PS/2 keyboard buffer
**************************************************************************************************
*/

/*
**************************************************************************************************
											GLOBAL VARIABLES
**************************************************************************************************
*/


INT8U	PS2KeyBuf[PS2KEY_BUF_SIZE];	//PS2 Keyboard buffer, the register to store characters key in
INT8U	PS2KeyBufInIx;				//Index into PS/2 key buf where next scan code will be inserted
INT8U	PS2KeyBufOutIx;				//Index into PS/2 key buf where next scan code will be removed
INT8U	PS2KeyNRead;				//Number of keys read from the PS/2 keyboard
INT8U	PS2Key_Code;				//Key code from PS/2 keyboard
INT8U	PS2Key_bit_Ix;				//bit index of the byte stream, Start_bit(1), Parity(10), Stop_bit(11)

INT8U	PS2STATUS;					//status of leds, ...
#bit PS2_SCROLL  = PS2STATUS.0		//TRUE if SCROLL LOCKed
#bit PS2_NUM     = PS2STATUS.1		//TRUE if NUM LOCKed
#bit PS2_CAPS    = PS2STATUS.2		//TRUE if CAPS LOCKed
#bit PS2_SHIFTED = PS2STATUS.3		//TRUE if SHIFTed
#bit PS2_CTRLED  = PS2STATUS.4		//TRUE if CTRLed
#bit PS2_ALTED   = PS2STATUS.5		//TRUE if ALTed
#bit PS2_BREAK   = PS2STATUS.6		//TRUE if a BREAKCODE received
#bit PS2_PRESENT = PS2STATUS.7		//TRUE if kbd connected

/*
**************************************************************************************************
									LOCAL FUNCTION PROTOTYPES
**************************************************************************************************
*/
void 	PS2KeyBufIn(INT8U code);	// Insert scan code into keyboard buffer
INT8U 	PS2KeyDecode(void);			// Function to decode a key stroke
void	PS2SetLeds(INT8U pattern);	// set NUM, SCROLL, CAPS KEDs 

/*
**************************************************************************************************
*								INSERT KEY CHARACTER INTO KEYBOARD BUFFER
* Description: This function inserts a key character into the keyboard buffer from a PS/2 keyboard
**************************************************************************************************
*/

void PS2KeyBufIn (INT8U code)
{
	if (PS2KeyNRead < PS2Key_BUF_SIZE) {		//make sure that we don't overflow the buffer
		PS2KeyNRead++;							//Increment the number of keys read
		PS2KeyBuf[PS2KeyBufInIx++] = code;		//Store the scan code into the buffer
		if (PS2KeyBufInIx >= PS2Key_BUF_SIZE) {	//Adjust index to the next scan code to put in buffer
			PS2KeyBufInIx = 0;
		}
	}
}

/*
**************************************************************************************************
*									DECODE PS/2 KEYBOARD SCAN CODE
* Description: This function is called to determine the key scane code of the PS/2 keyboard
* Arguments:   none
* Returns:	   the key scan code
**************************************************************************************************
*/
INT8U PS2KeyDecode(char c)
{	
	return (PS2_CAPS || PS2_SHIFTED) ? PS2KeyShifted[c] : PS2KeyUnshifted[c];
}

/*
**************************************************************************************************
*										GET KEY	
* Description: 	Get a PS/2 keyboard scan code from the keyboard driver
* Arguments:   	none
* Returns:	   	!=0xFF is the key scan code of the PS/2 keyboard key pressed
*				==0xFF indicates that there is no key in the PS2KeyBuf[PS2KEY_BUF_SIZE] buffer 
**************************************************************************************************
*/

INT8U PS2KeyGetKey (void)
{
	INT8U code;

	if (PS2KeyNRead > 0) {					/* See if we have keys in PS2KeyBuf[] */
		PS2KeyNRead--;						/* Decrement the number of keys in the buffer */
		code = PS2KeyBuf[PS2KeyBufOutIx];	/* Get scan code from the buffer */
		PS2KeyBufOutIx++;
		if (PS2KeyBufOutIx >= PS2Key_BUF_SIZE) {
			PS2KeyBufOutIx = 0;
		}
		
		// buffer !full => release clk line
		if (PS2KeyNRead < PS2Key_BUF_SIZE){
			PS2Key_ClkTris = 1;
			output_high(PS2Key_Clk);
		}	
		return (code);						/* return the scan code of the key pressed */
	}
	else {
		return (0xFF);						/* No scan codes in buffer, return -1 */
	}
}

/*
**************************************************************************************************
*										SEND COMMAND
* Description: 	Send command to the PS/2 keyboard
* Arguments:   	cmd is the command to send
*				$ED Set Status LED's - This command can be used to turn on and off the Num Lock, Caps Lock & Scroll Lock LED's. 
*					After Sending ED, keyboard will reply with ACK (FA) and wait for another byte which determines their Status.
*					Bit 0 controls the Scroll Lock, Bit 1 the Num Lock and Bit 2 the Caps lock. Bits 3 to 7 are ignored.  
* 				$EE Echo - Upon sending a Echo command to the Keyboard, the keyboard should reply with a Echo (EE) 
*				$F0 Set Scan Code Set. Upon Sending F0, keyboard will reply with ACK (FA) and wait for another byte, 
					01-03 which determines the Scan Code Used. 
					Sending 00 as the second byte will return the Scan Code Set currently in Use
*				$F3 Set Typematic Repeat Rate. Keyboard will Acknowledge command with FA and wait for second byte, 
					which determines the Typematic Repeat Rate. 
*				$F4 Keyboard Enable - Clears the keyboards output buffer, enables Keyboard Scanning and returns an Acknowledgment. 
*				$F5 Keyboard Disable - Resets the keyboard, disables Keyboard Scanning and returns an Acknowledgment. 
*				$FE Resend - Upon receipt of the resend command the keyboard will re- transmit the last byte sent. 
*				$FF Reset - Resets the Keyboard. 

* Returns:	   	none 
* Remarks:      timing was a little bit tricky, but works with my noname keyboard
*               
**************************************************************************************************
*/

void PS2SendComm(INT8U cmd)
{
	unsigned int16 tmp;

	PS2Key_bit_Ix = 0;				//bit index of the byte stream, Start_bit(1), Parity(10), Stop_bit(11), ACK_bit(12)

	// calculate odd parity, shift to high byte, or with command
	tmp = ((((((cmd >> 4) | (cmd << 4)) ^ cmd) + 0X41) & 0X82) + 0X7E) & 0X80;
	tmp = ((tmp<<1) ^ 0xFF00) | cmd; 	

	delay_ms(5);
	disable_interrupts(INT_EXT2);	//disable external interrupt B2

	PS2Key_ClkTris = 0;
	output_low(PS2Key_Clk); 		//host forces PS2 clock line low.
	delay_us(120);					//After approx. 120us, the host pulls PS2 data line low

	PS2Key_DataTris = 0;
	output_low(PS2Key_Data);		//This sequence of event signals the keyboard that 
	delay_us(5);					//the host pulls PS2 data line low
									//the host is about to transfer a command
	output_high(PS2Key_Clk); 		//host forces PS2 clock line low.
	PS2Key_ClkTris = 1;				//resistor after the falling edge of the data signal
	delay_us(5);

	/* Transfer of data from host to keyboard will start approx. 1ms after the rising edge of the clock */
	/* shift bit to PS2Key_Data on every falling edge of the PS2Key_Clk line */
	
	
	do {
		while (input(PS2Key_Clk))	// wait for H->L
		delay_us(5);
		output_bit(PS2Key_Data, tmp&0x01);
		while (!input(PS2Key_Clk));	// wait for L->H
		delay_us(5);
		tmp = tmp>>1;		
		PS2Key_bit_Ix++;
		}
	while(PS2Key_bit_Ix<=10);
	
	output_high(PS2Key_Data);
	PS2Key_DataTris = 1;			//release data line, reset to input
	
	output_low(PS2Key_Clk);			// bus inhibit! my kbd is to fast
	PS2Key_bit_Ix = 0;
	//delay_us(100);
	
	ext_int_edge(2, H_TO_L);		//set external interrupt to falling edge
	output_high(PS2Key_Clk);		// bus enable
	
	PS2Key_ClkTris = 1;
 	enable_interrupts(INT_EXT2);
}

/*
**************************************************************************************************
*									SEE IF ANY KEY IN BUFFER	
* Description: 	This function checks to see if a key was pressed
* Arguments:   	none
* Returns:	   	TRUE if a key has been pressed
*				FALSE if no key pressed
**************************************************************************************************
*/

BOOLEAN PS2KeyHit(void)
{
	BOOLEAN hit;
	hit = (BOOLEAN)(PS2KeyNRead > 0)? TRUE : FALSE;
	return(hit);
}

/*
**************************************************************************************************
*									SEE IF PS2-Keyboard connected	
* Description: 	This function checks if there is a PS2 keyboard connected
* Arguments:   	none
* Returns:	   	TRUE if available
*				FALSE if not
**************************************************************************************************
*/
BOOLEAN PS2TestKeyboard(void){
	int i;
	BOOLEAN KbdConnected = FALSE;
	disable_interrupts(INT_EXT2);

	delay_ms(1000);
	
	// Clock L=>H + Data low => kbd sends clock pulses 
	PS2Key_ClkTris = 0;
	output_low(PS2Key_Clk);
	PS2Key_DataTris = 0;
	output_low(PS2Key_Data);

	delay_us(120);

	output_high(PS2Key_Clk);
	PS2Key_ClkTris = 1;
	delay_us(5);
	
	// look for clock pulses
	for (i=0; i<100; i++){
		delay_us(10);
		KbdConnected |= !input(PS2Key_Clk);
	}

	PS2_PRESENT=KbdConnected;
		
	// bus idle
	PS2Key_DataTris = 1;
	output_high(PS2Key_Data);

	delay_ms(10);
	enable_interrupts(INT_EXT2);
	return KbdConnected;
}



/*
*********************************************************************************************************
*                                  PS2 KEYBOARD DRIVER INITIALIZATION
*
* Description : This function initializes the PS2 driver. Quick & very dirty
* Arguments   : None.
* Returns     : None.
*********************************************************************************************************
*/

void PS2KeyInit(void)
{
	PS2KeyNRead 	=0;				//Clear the number of keys read
	PS2KeyBufInIx 	=0;				//Clear the key codes inserted at the beginning of the buffer
	PS2KeyBufOutIx	=0;				//Clear the key codes removed at the beginning of the buffer
	PS2_BREAK		=FALSE;

	PS2_SHIFTED		=FALSE;
	PS2_SCROLL		=FALSE;
	PS2_NUM			=FALSE;
	PS2_CAPS		=FALSE;
	PS2_CTRLED		=FALSE;
	PS2_ALTED		=FALSE;

	PS2Key_DataTris =1;				//Set Data line an input
	PS2Key_ClkTris	=1;				//Set clock line an input

	ext_int_edge(2, H_TO_L);		//set external interrupt to falling edge
	enable_interrupts(INT_EXT2);	//enable external interrupt B2
	enable_interrupts(GLOBAL);
}




/*
*********************************************************************************************************
*                                     PS2 KEYBOARD INTERRUPT HANDLER
* Description:  This function is called by an interrupt handler (e.g. B0 EXT INT on falling edge)
* 				Detect the clock pulse (High_to_Low) for new bit
*				Total 11 bits for a complete frame (keyboard -> Host)
* Arguments  : 	none
* Returns    : 	none
* Notes      : 	Put a scan code to PS2KeyBuf[] if a valid byte acquired
*				Set PS2KeyNRead then
*********************************************************************************************************
*/

void PS2KeyIntHandler(void)
{
	BOOLEAN	PS2Key_Data_State;	//Temp variable for Data pin, PS2Key_Data_State
	BOOLEAN PS2Key_Clk_State;
	BOOLEAN	PS2Status_Parity;	//Parity bit	
		
	PS2Key_Data_State = input(PS2Key_Data);
	PS2Key_Clk_State  = input(PS2Key_Clk);
	PS2Key_bit_Ix++;

	if (PS2Key_bit_Ix==PS2Key_START) {
		//  First bit, PS2Key_START = 1 as defined under ps2key.h
      	if (PS2Key_Data_State || PS2Key_Clk_State)	PS2Key_bit_Ix=0;
		//	Start bit clock is a high_to_low transition.
		//	Start bit is a low bit. If data pin or Clk pin high for the Start bit,
		//  it must be a false trigger from RB0, thus setting PS2Key_bit_Ix=0
      	PS2Key_Code=0;				
      	PS2Status_Parity=0;
	} 
	
	else if (PS2Key_bit_Ix==PS2Key_PARITY)
	{		//  Parity bit, PS2Key_PARITY = 10 as defined under ps2key.h
			PS2Status_Parity = PS2Key_Data_State;
	}

	// key received and ready
	else if (PS2Key_bit_Ix==PS2Key_STOP)
	{	// Stop bit (PS2Key_STOP = 11)
       	PS2Key_bit_Ix=0;
       						
		switch (PS2Key_Code) {
			case 0xF0: // BREAK recieved
				PS2_BREAK = TRUE;
				break;
				
			case 0xFA: // ACK recieved
				break;
				
			case 0xAA: // kbd POST ok
				PS2_PRESENT=TRUE;
				break;

			// SHIFT
			case 0x12:	// left  SHIFT
			case 0x59:	// right SHIFT
				if (PS2_BREAK){
					PS2_SHIFTED = FALSE;
					PS2_BREAK = FALSE;
				}
				else {
					PS2_SHIFTED = TRUE;
				}		
				break;

			// CTRL
			case 0x14:	// left  CTRL
			//case 0x59:	// right CTRL
				if (PS2_BREAK){
					PS2_CTRLED = FALSE;
					PS2_BREAK = FALSE;
				}
				else {
					PS2_CTRLED = TRUE;
				}		
				break;

			// ALT
			case 0x11:	// left ALT
				if (PS2_BREAK){
					PS2_ALTED = FALSE;
					PS2_BREAK = FALSE;
				}
				else {
					PS2_ALTED = TRUE;
				}		
				break;
			
			case 0x58:	// CAPS Lock
				if (PS2_BREAK){
					PS2_CAPS = !PS2_CAPS;
					PS2_BREAK = FALSE;
					PS2SendComm(0xED);
					PS2SendComm(PS2STATUS & 0x07);
				}
				break;

			case 0x77:	// NUM Lock
				if (PS2_BREAK){
					PS2_NUM = !PS2_NUM;
					PS2_BREAK = FALSE;
					PS2SendComm(0xED);
					PS2SendComm(PS2STATUS & 0x07);
				}
				break;			

			case 0x7E:	// scroll = reset kbd  <===============
				PS2SendComm(0xFF);
				delay_ms(1500);
				break;

			default:
				if (PS2_BREAK) {
					// Put valid byte in PS2KeyBuf[]
					PS2KeyBufIn (PS2Key_Code);
	
					// buffer full? => clk low = inhibit
					if (PS2KeyNRead >= PS2Key_BUF_SIZE) {
						delay_us(30);
						PS2Key_ClkTris = 0;
						output_low(PS2Key_Clk);	
					}
					
					PS2_PRESENT=TRUE;
					PS2_BREAK = FALSE;
					break;
				}
		}
	}

	// shift in 8 bit
	else
	{		// For all bits from 2, 3...9 (total 8-bit)
       		PS2Key_Code = PS2Key_Code >> 1;
      		    if (PS2Key_Data_State) 
				PS2Key_Code = PS2Key_Code | 0x80;
 		}
}



// some tools, not yet finished
void PS2SetLeds(INT8U pattern){
	PS2SendComm(0xED);
	PS2SendComm(pattern & 0x07);
}

